# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import gcs_service_pb2 as src_dot_ray_dot_protobuf_dot_gcs__service__pb2


class JobInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.AddJob = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/AddJob',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddJobRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddJobReply.FromString,
        )
    self.MarkJobFinished = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/MarkJobFinished',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.MarkJobFinishedRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.MarkJobFinishedReply.FromString,
        )
    self.GetAllJobInfo = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/GetAllJobInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllJobInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllJobInfoReply.FromString,
        )
    self.ReportJobError = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/ReportJobError',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportJobErrorRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportJobErrorReply.FromString,
        )
    self.GetNextJobID = channel.unary_unary(
        '/ray.rpc.JobInfoGcsService/GetNextJobID',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNextJobIDRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNextJobIDReply.FromString,
        )


class JobInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def AddJob(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def MarkJobFinished(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllJobInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ReportJobError(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNextJobID(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_JobInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'AddJob': grpc.unary_unary_rpc_method_handler(
          servicer.AddJob,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddJobRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddJobReply.SerializeToString,
      ),
      'MarkJobFinished': grpc.unary_unary_rpc_method_handler(
          servicer.MarkJobFinished,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.MarkJobFinishedRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.MarkJobFinishedReply.SerializeToString,
      ),
      'GetAllJobInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllJobInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllJobInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllJobInfoReply.SerializeToString,
      ),
      'ReportJobError': grpc.unary_unary_rpc_method_handler(
          servicer.ReportJobError,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportJobErrorRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportJobErrorReply.SerializeToString,
      ),
      'GetNextJobID': grpc.unary_unary_rpc_method_handler(
          servicer.GetNextJobID,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNextJobIDRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNextJobIDReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.JobInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class ActorInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.RegisterActor = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/RegisterActor',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterActorRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterActorReply.FromString,
        )
    self.CreateActor = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/CreateActor',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreateActorRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreateActorReply.FromString,
        )
    self.GetActorInfo = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/GetActorInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetActorInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetActorInfoReply.FromString,
        )
    self.GetNamedActorInfo = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/GetNamedActorInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedActorInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedActorInfoReply.FromString,
        )
    self.GetAllActorInfo = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/GetAllActorInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllActorInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllActorInfoReply.FromString,
        )
    self.KillActorViaGcs = channel.unary_unary(
        '/ray.rpc.ActorInfoGcsService/KillActorViaGcs',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.KillActorViaGcsRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.KillActorViaGcsReply.FromString,
        )


class ActorInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def RegisterActor(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateActor(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetActorInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNamedActorInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllActorInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def KillActorViaGcs(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ActorInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'RegisterActor': grpc.unary_unary_rpc_method_handler(
          servicer.RegisterActor,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterActorRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterActorReply.SerializeToString,
      ),
      'CreateActor': grpc.unary_unary_rpc_method_handler(
          servicer.CreateActor,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreateActorRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreateActorReply.SerializeToString,
      ),
      'GetActorInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetActorInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetActorInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetActorInfoReply.SerializeToString,
      ),
      'GetNamedActorInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetNamedActorInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedActorInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedActorInfoReply.SerializeToString,
      ),
      'GetAllActorInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllActorInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllActorInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllActorInfoReply.SerializeToString,
      ),
      'KillActorViaGcs': grpc.unary_unary_rpc_method_handler(
          servicer.KillActorViaGcs,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.KillActorViaGcsRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.KillActorViaGcsReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.ActorInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class NodeInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.RegisterNode = channel.unary_unary(
        '/ray.rpc.NodeInfoGcsService/RegisterNode',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterNodeRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterNodeReply.FromString,
        )
    self.UnregisterNode = channel.unary_unary(
        '/ray.rpc.NodeInfoGcsService/UnregisterNode',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UnregisterNodeRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UnregisterNodeReply.FromString,
        )
    self.GetAllNodeInfo = channel.unary_unary(
        '/ray.rpc.NodeInfoGcsService/GetAllNodeInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllNodeInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllNodeInfoReply.FromString,
        )
    self.GetInternalConfig = channel.unary_unary(
        '/ray.rpc.NodeInfoGcsService/GetInternalConfig',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetInternalConfigRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetInternalConfigReply.FromString,
        )


class NodeInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def RegisterNode(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UnregisterNode(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllNodeInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetInternalConfig(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_NodeInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'RegisterNode': grpc.unary_unary_rpc_method_handler(
          servicer.RegisterNode,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterNodeRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RegisterNodeReply.SerializeToString,
      ),
      'UnregisterNode': grpc.unary_unary_rpc_method_handler(
          servicer.UnregisterNode,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UnregisterNodeRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UnregisterNodeReply.SerializeToString,
      ),
      'GetAllNodeInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllNodeInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllNodeInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllNodeInfoReply.SerializeToString,
      ),
      'GetInternalConfig': grpc.unary_unary_rpc_method_handler(
          servicer.GetInternalConfig,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetInternalConfigRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetInternalConfigReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.NodeInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class HeartbeatInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ReportHeartbeat = channel.unary_unary(
        '/ray.rpc.HeartbeatInfoGcsService/ReportHeartbeat',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportHeartbeatRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportHeartbeatReply.FromString,
        )


class HeartbeatInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def ReportHeartbeat(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_HeartbeatInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ReportHeartbeat': grpc.unary_unary_rpc_method_handler(
          servicer.ReportHeartbeat,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportHeartbeatRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportHeartbeatReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.HeartbeatInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class ObjectInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetObjectLocations = channel.unary_unary(
        '/ray.rpc.ObjectInfoGcsService/GetObjectLocations',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetObjectLocationsRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetObjectLocationsReply.FromString,
        )
    self.GetAllObjectLocations = channel.unary_unary(
        '/ray.rpc.ObjectInfoGcsService/GetAllObjectLocations',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllObjectLocationsRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllObjectLocationsReply.FromString,
        )
    self.AddObjectLocation = channel.unary_unary(
        '/ray.rpc.ObjectInfoGcsService/AddObjectLocation',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddObjectLocationRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddObjectLocationReply.FromString,
        )
    self.RemoveObjectLocation = channel.unary_unary(
        '/ray.rpc.ObjectInfoGcsService/RemoveObjectLocation',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemoveObjectLocationRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemoveObjectLocationReply.FromString,
        )


class ObjectInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def GetObjectLocations(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllObjectLocations(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddObjectLocation(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RemoveObjectLocation(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ObjectInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetObjectLocations': grpc.unary_unary_rpc_method_handler(
          servicer.GetObjectLocations,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetObjectLocationsRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetObjectLocationsReply.SerializeToString,
      ),
      'GetAllObjectLocations': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllObjectLocations,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllObjectLocationsRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllObjectLocationsReply.SerializeToString,
      ),
      'AddObjectLocation': grpc.unary_unary_rpc_method_handler(
          servicer.AddObjectLocation,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddObjectLocationRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddObjectLocationReply.SerializeToString,
      ),
      'RemoveObjectLocation': grpc.unary_unary_rpc_method_handler(
          servicer.RemoveObjectLocation,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemoveObjectLocationRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemoveObjectLocationReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.ObjectInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class TaskInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.AddTask = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/AddTask',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskReply.FromString,
        )
    self.GetTask = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/GetTask',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskReply.FromString,
        )
    self.AddTaskLease = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/AddTaskLease',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskLeaseRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskLeaseReply.FromString,
        )
    self.GetTaskLease = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/GetTaskLease',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskLeaseRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskLeaseReply.FromString,
        )
    self.AttemptTaskReconstruction = channel.unary_unary(
        '/ray.rpc.TaskInfoGcsService/AttemptTaskReconstruction',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AttemptTaskReconstructionRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AttemptTaskReconstructionReply.FromString,
        )


class TaskInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def AddTask(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTask(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddTaskLease(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTaskLease(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AttemptTaskReconstruction(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_TaskInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'AddTask': grpc.unary_unary_rpc_method_handler(
          servicer.AddTask,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskReply.SerializeToString,
      ),
      'GetTask': grpc.unary_unary_rpc_method_handler(
          servicer.GetTask,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskReply.SerializeToString,
      ),
      'AddTaskLease': grpc.unary_unary_rpc_method_handler(
          servicer.AddTaskLease,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskLeaseRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddTaskLeaseReply.SerializeToString,
      ),
      'GetTaskLease': grpc.unary_unary_rpc_method_handler(
          servicer.GetTaskLease,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskLeaseRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetTaskLeaseReply.SerializeToString,
      ),
      'AttemptTaskReconstruction': grpc.unary_unary_rpc_method_handler(
          servicer.AttemptTaskReconstruction,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AttemptTaskReconstructionRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AttemptTaskReconstructionReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.TaskInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class StatsGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.AddProfileData = channel.unary_unary(
        '/ray.rpc.StatsGcsService/AddProfileData',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddProfileDataRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddProfileDataReply.FromString,
        )
    self.GetAllProfileInfo = channel.unary_unary(
        '/ray.rpc.StatsGcsService/GetAllProfileInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllProfileInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllProfileInfoReply.FromString,
        )


class StatsGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def AddProfileData(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllProfileInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_StatsGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'AddProfileData': grpc.unary_unary_rpc_method_handler(
          servicer.AddProfileData,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddProfileDataRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddProfileDataReply.SerializeToString,
      ),
      'GetAllProfileInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllProfileInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllProfileInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllProfileInfoReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.StatsGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class WorkerInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ReportWorkerFailure = channel.unary_unary(
        '/ray.rpc.WorkerInfoGcsService/ReportWorkerFailure',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportWorkerFailureRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportWorkerFailureReply.FromString,
        )
    self.GetWorkerInfo = channel.unary_unary(
        '/ray.rpc.WorkerInfoGcsService/GetWorkerInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetWorkerInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetWorkerInfoReply.FromString,
        )
    self.GetAllWorkerInfo = channel.unary_unary(
        '/ray.rpc.WorkerInfoGcsService/GetAllWorkerInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllWorkerInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllWorkerInfoReply.FromString,
        )
    self.AddWorkerInfo = channel.unary_unary(
        '/ray.rpc.WorkerInfoGcsService/AddWorkerInfo',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddWorkerInfoRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddWorkerInfoReply.FromString,
        )


class WorkerInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def ReportWorkerFailure(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetWorkerInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllWorkerInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddWorkerInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_WorkerInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ReportWorkerFailure': grpc.unary_unary_rpc_method_handler(
          servicer.ReportWorkerFailure,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportWorkerFailureRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportWorkerFailureReply.SerializeToString,
      ),
      'GetWorkerInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetWorkerInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetWorkerInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetWorkerInfoReply.SerializeToString,
      ),
      'GetAllWorkerInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllWorkerInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllWorkerInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllWorkerInfoReply.SerializeToString,
      ),
      'AddWorkerInfo': grpc.unary_unary_rpc_method_handler(
          servicer.AddWorkerInfo,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddWorkerInfoRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.AddWorkerInfoReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.WorkerInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class PlacementGroupInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.CreatePlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/CreatePlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreatePlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreatePlacementGroupReply.FromString,
        )
    self.RemovePlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/RemovePlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemovePlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemovePlacementGroupReply.FromString,
        )
    self.GetPlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/GetPlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetPlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetPlacementGroupReply.FromString,
        )
    self.GetNamedPlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/GetNamedPlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedPlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedPlacementGroupReply.FromString,
        )
    self.GetAllPlacementGroup = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/GetAllPlacementGroup',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllPlacementGroupRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllPlacementGroupReply.FromString,
        )
    self.WaitPlacementGroupUntilReady = channel.unary_unary(
        '/ray.rpc.PlacementGroupInfoGcsService/WaitPlacementGroupUntilReady',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.WaitPlacementGroupUntilReadyRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.WaitPlacementGroupUntilReadyReply.FromString,
        )


class PlacementGroupInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def CreatePlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RemovePlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetPlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNamedPlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllPlacementGroup(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def WaitPlacementGroupUntilReady(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PlacementGroupInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'CreatePlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.CreatePlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreatePlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.CreatePlacementGroupReply.SerializeToString,
      ),
      'RemovePlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.RemovePlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemovePlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.RemovePlacementGroupReply.SerializeToString,
      ),
      'GetPlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.GetPlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetPlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetPlacementGroupReply.SerializeToString,
      ),
      'GetNamedPlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.GetNamedPlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedPlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetNamedPlacementGroupReply.SerializeToString,
      ),
      'GetAllPlacementGroup': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllPlacementGroup,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllPlacementGroupRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllPlacementGroupReply.SerializeToString,
      ),
      'WaitPlacementGroupUntilReady': grpc.unary_unary_rpc_method_handler(
          servicer.WaitPlacementGroupUntilReady,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.WaitPlacementGroupUntilReadyRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.WaitPlacementGroupUntilReadyReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.PlacementGroupInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class InternalKVGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.InternalKVGet = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVGet',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVGetRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVGetReply.FromString,
        )
    self.InternalKVPut = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVPut',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVPutRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVPutReply.FromString,
        )
    self.InternalKVDel = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVDel',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVDelRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVDelReply.FromString,
        )
    self.InternalKVExists = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVExists',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVExistsRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVExistsReply.FromString,
        )
    self.InternalKVKeys = channel.unary_unary(
        '/ray.rpc.InternalKVGcsService/InternalKVKeys',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVKeysRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVKeysReply.FromString,
        )


class InternalKVGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def InternalKVGet(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def InternalKVPut(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def InternalKVDel(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def InternalKVExists(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def InternalKVKeys(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_InternalKVGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'InternalKVGet': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVGet,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVGetRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVGetReply.SerializeToString,
      ),
      'InternalKVPut': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVPut,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVPutRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVPutReply.SerializeToString,
      ),
      'InternalKVDel': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVDel,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVDelRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVDelReply.SerializeToString,
      ),
      'InternalKVExists': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVExists,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVExistsRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVExistsReply.SerializeToString,
      ),
      'InternalKVKeys': grpc.unary_unary_rpc_method_handler(
          servicer.InternalKVKeys,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVKeysRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.InternalKVKeysReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.InternalKVGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class NodeResourceInfoGcsServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetResources = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/GetResources',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetResourcesRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetResourcesReply.FromString,
        )
    self.UpdateResources = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/UpdateResources',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UpdateResourcesRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UpdateResourcesReply.FromString,
        )
    self.DeleteResources = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/DeleteResources',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.DeleteResourcesRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.DeleteResourcesReply.FromString,
        )
    self.GetAllAvailableResources = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/GetAllAvailableResources',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllAvailableResourcesRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllAvailableResourcesReply.FromString,
        )
    self.ReportResourceUsage = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/ReportResourceUsage',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportResourceUsageRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportResourceUsageReply.FromString,
        )
    self.GetAllResourceUsage = channel.unary_unary(
        '/ray.rpc.NodeResourceInfoGcsService/GetAllResourceUsage',
        request_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllResourceUsageRequest.SerializeToString,
        response_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllResourceUsageReply.FromString,
        )


class NodeResourceInfoGcsServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def GetResources(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UpdateResources(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteResources(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllAvailableResources(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ReportResourceUsage(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAllResourceUsage(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_NodeResourceInfoGcsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetResources': grpc.unary_unary_rpc_method_handler(
          servicer.GetResources,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetResourcesRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetResourcesReply.SerializeToString,
      ),
      'UpdateResources': grpc.unary_unary_rpc_method_handler(
          servicer.UpdateResources,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UpdateResourcesRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.UpdateResourcesReply.SerializeToString,
      ),
      'DeleteResources': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteResources,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.DeleteResourcesRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.DeleteResourcesReply.SerializeToString,
      ),
      'GetAllAvailableResources': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllAvailableResources,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllAvailableResourcesRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllAvailableResourcesReply.SerializeToString,
      ),
      'ReportResourceUsage': grpc.unary_unary_rpc_method_handler(
          servicer.ReportResourceUsage,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportResourceUsageRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.ReportResourceUsageReply.SerializeToString,
      ),
      'GetAllResourceUsage': grpc.unary_unary_rpc_method_handler(
          servicer.GetAllResourceUsage,
          request_deserializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllResourceUsageRequest.FromString,
          response_serializer=src_dot_ray_dot_protobuf_dot_gcs__service__pb2.GetAllResourceUsageReply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ray.rpc.NodeResourceInfoGcsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))
